package TwoStageCPU ( mkTwoStageCPU ) where

import FIFO
import Array

data RName = R0  | R1  | R2  | R3  | R4  | R5  | R6  | R7
	   | R8  | R9  | R10 | R11 | R12 | R13 | R14 | R15
	   | R16 | R17 | R18 | R19 | R20 | R21 | R22 | R23
	   | R24 | R25 | R26 | R27 | R28 | R29 | R30 | R31
    deriving (Eq, Bits, Bounded)

type Src   = RName
type Dest  = RName
type Cond  = RName
type Addr  = RName
type Val   = RName

type IAddr = Bit 32   -- instruction memory address
type MVal  = Bit 32   -- memory value

data Instr = Add   Dest Src Src
	   | Bz    Cond Addr
	   | Load  Dest Addr
	   | Store Val  Addr
    deriving (Eq, Bits)

-- --------------------------------------------------------
-- Register File

interface RegFile =
    read  :: RName -> MVal
    write :: RName -> MVal -> Action

mkRegFile :: Module RegFile
mkRegFile =
    module
	-- r131 :: Array RName MVal <- mkArray R1 R31
	r131 :: Array RName MVal <- mkArray R0 R31
        interface
	    read  addr = if (addr == R0) then 0 else r131.sub addr
	    write addr val = r131.upd addr val

interface MemIF =
    read  :: (Bit 32) -> (Bit 32)
    write :: (Bit 32) -> (Bit 32) -> Action

dummyMem :: MemIF
dummyMem = interface MemIF
               read  a   = a
               write a v = action

-- -------------------------------------------------------
-- The Unpipelined Processor

mkTwoStageCPU :: Module Empty
mkTwoStageCPU =
    module
	let imem = dummyMem
	    dmem = dummyMem

	pc :: Reg IAddr <- mkReg 0
	rf :: RegFile <- mkRegFile
	bu :: FIFO Instr <- mkFIFO

	rules

	-- Fetch Rules --

	  "Fetch":
	    when True
	      ==> action  pc := pc + 1
		          bu.enq (unpack (truncate (imem.read pc)))

	-- Execute Rules --

	  "Add":
	    when (Add rd rs rt) <- bu.first
	      ==> action  rf.write rd (rf.read rs + rf.read rt)
		          bu.deq

	  "Bz Not Taken":
	    when (Bz rc ra) <- bu.first, (rf.read rc) /= 0
	      ==> action  bu.deq

	  "Bz Taken":
	    when (Bz rc ra) <- bu.first, (rf.read rc) == 0
	      ==> action  pc := rf.read ra
		          bu.clear

